"""Various icons which can be used in the application."""

from __future__ import annotations

FAVORITE_EMOJI = "🌟"
FOLDER_EMOJI = "📁"
FOLDER_OPEN_EMOJI = "📂"
GLOBE_EMOJI = "🌎"
NOTE_EMOJI = "📄"
CREDIT_CARD_EMOJI = "💳"
ID_CARD_EMOJI = "🪪"
TRASH_EMOJI = "🗑"
CLIPBOARD_EMOJI = "📋"
PENCIL_EMOJI = "✏"
MEMO_EMOJI = "📝"
PAPER_EMOJI = "📄"
BLANK_PAGE_EMOJI = "📃"
COPY_EMOJI = "❐"
GEAR_EMOJI = "⚙"
EYE_EMOJI = "👁"
LINK_EMOJI = "🔗"
HEAVY_PLUS_SIGN_EMOJI = "➕"
CROSS_MARK_EMOJI = "❌"
KEY_EMOJI = "🔑"
PERSON_EMOJI = "👤"
UNKNOWN_EMOJI = "❓"
DOWN_CHEVRON_EMOJI = "🔽"
